dinvgamma <- function(x, shape, scale = 1, log=FALSE) {
    if (any(shape <= 0 | scale <= 0)) {
        stop("Shape or scale parameter negative in dinvgamma().\n")
    }
    alpha <- shape
    beta <- scale
    log.density <- alpha * log(beta) - lgamma(alpha) - (alpha + 
        1) * log(x) - (beta/x)
    ifelse(log, log.density, exp(log.density))
}

log.sum.exp <- function(x) {
  m <- max(x)
  log(sum(exp(x - m))) + m
}

eval.log.density <- function(pt, clusters)
  colSums(
          matrix(dnorm(pt,
                       mean=t(clusters$mean),
                       sd=t(clusters$sd),
                       log=TRUE),
                 ncol=nrow(clusters$mean))
          ) + clusters$log.proportions


generate.hyper.params <- function(points, clusters.num) {
 list(mu=colMeans(points),
      kappa=rep(0.01, ncol(points)),
      nu=rep(3, ncol(points)),
      ksi=pmax(1e-3, apply(points, 2, sd)/clusters.num))
}

next.clusters <- function(points, clusters, sd.modes, hyper.params) {
  points.num <- nrow(points)
  clusters.num <- nrow(clusters$mean)
  f <- function(pt) # evaluate and sum all densities at a given point (in log-space)
    log.sum.exp(eval.log.density(pt, clusters))
  log.normalizing.constants <- apply(points, 1, f)
  result <- clusters
  g <- function(k) { # evaluate a given density at every point; find new cluster parameters
    log.weights <- colSums(
                           matrix(dnorm(t(points),
                                        mean=clusters$mean[k,],
                                        sd=clusters$sd[k,],
                                        log=TRUE),
                                  ncol=points.num)
                           ) + clusters$log.proportions[k] - log.normalizing.constants  # this is log z[,k] from tr486.pdf
    v <- log.sum.exp(log.weights) # this is log n[k] from tr486.pdf
    w <- exp(log.weights - v) # sum(w) == 1 holds
    result$mean[k,] <<- (exp(v) * colSums(points * w) + hyper.params$kappa * hyper.params$mu) / (hyper.params$kappa + exp(v))
#    result$mean[k,] <<- colSums(points * w)
    result$log.proportions[k] <<- v - log(points.num)
    mask <- which(sd.modes==2)
    if(length(mask) > 0)
      result$sd[k,mask] <<- sqrt((hyper.params$ksi[mask]^2 + hyper.params$kappa[mask] * exp(v) / (hyper.params$kappa[mask] + exp(v)) * (colSums(points[,mask,drop=FALSE] * w) - hyper.params$mu[mask])^2 + exp(v) * colSums(t((t(points[,mask,drop=FALSE]) - result$mean[k,mask])^2) * w)) / (hyper.params$nu[mask] + exp(v) + 2))
#      result$sd[k,mask] <<- sqrt(colSums(t((t(points[,mask]) - result$mean[k,mask])^2) * w))
    mask <- which(sd.modes==3)
    if(length(mask) > 0)
      result$sd[k,mask] <<- hyper.params$kappa[mask] * exp(v) / (hyper.params$kappa[mask] + exp(v)) * (colSums(points[,mask,drop=FALSE] * w) - hyper.params$mu[mask])^2 + colSums(t((t(points[,mask,drop=FALSE]) - result$mean[k,mask])^2) * exp(log.weights))
#      result$sd[k,mask] <<- colSums(t((t(points[,mask]) - result$mean[k,mask])^2) * exp(log.weights))
  }
  t(sapply(1:clusters.num, g))
  mask <- which(sd.modes==3)
  if(length(mask) > 0)
    result$sd[,mask] <- t(replicate(clusters.num, sqrt((hyper.params$ksi[mask]^2 + colSums(result$sd[,mask,drop=FALSE])) / (hyper.params$nu[mask] + points.num + 2))))
  result
}

default.eps <- list(mean=1e-6, sd=1e-6, proportions=1e-6)

next.clusters.until.converged <- function(points, clusters, sd.modes, hyper.params, eps=default.eps) {
  repeat {
    prev.clusters <- clusters
    clusters <- next.clusters(points, clusters, sd.modes, hyper.params)
    if(max(rowSums((clusters$mean - prev.clusters$mean)^2)) < eps$mean &&
       max(rowSums((clusters$sd - prev.clusters$sd)^2)) < eps$sd &&
       max(sum((exp(clusters$log.proportions)-exp(clusters$log.proportions))^2)) < eps$proportions)
      break
  }
  clusters
}

best.assignment <- function(points, clusters) {
  f <- function(pt) {
    d <- eval.log.density(pt, clusters)
    i <- which.max(d)
    print(i)
    c(i, exp(d[i] - log.sum.exp(d)))
  }
  res <- t(apply(points, 1, f))
  colnames(res) <- c("cluster", "probability")
  res
}

log.likelihood <- function(points, clusters) {
  f <- function(pt)
    log.sum.exp(eval.log.density(pt, clusters))
  sum(apply(points, 1, f))
}

log.prior <- function(clusters, sd.modes, hyper.params) {
  res <- sum(dnorm(t(clusters$mean),
                   mean=hyper.params$mu,
                   sd=t(clusters$sd)/sqrt(hyper.params$kappa),
                   log=TRUE))
  mask <- which(sd.modes==2)
  if(length(mask) > 0)
    res <- res + sum(dinvgamma(t(clusters$sd[,mask])^2,
                               shape=0.5*hyper.params$nu[mask],
                               scale=0.5*hyper.params$ksi[mask]^2,
                               log=TRUE))
  mask <- which(sd.modes==3)
  if(length(mask) > 0)
    res <- res + sum(dinvgamma(t(clusters$sd[1,mask])^2,
                               shape=0.5*hyper.params$nu[mask],
                               scale=0.5*hyper.params$ksi[mask]^2,
                               log=TRUE))
  res
}

log.posterior.unnorm <- function(points, clusters, sd.modes, hyper.params)
  log.likelihood(points, clusters) + log.prior(clusters, sd.modes, hyper.params)
